;+
; NAME: FASTPHOT
;
; PURPOSE: Perform the simultaneous PSF fitting of sources at known position.
;
; CALLING SEQUENCE: fastphot, map, psf, noise_map, X, Y, flux, $
;                        Fprior = Fprior, map_clean = map_clean
;
; INPUTS:
;         MAP: Array containing the map of the studied field (all
;              the good pixel must be >-99998). 
;              If bad pixel, put -99999. value
;         PSF: Array containing the PSF. It must be the (2N+1)*(2N+1)
;         image (in the unit of the map) for a source with flux = 1
;         (in output unit).
;         NOISE_MAP: Array containing the standard deviation of the
;         noise. (same size as map) 
;         X: Vector containing the x coordinates of the sources. (float if decimal coordinates or not)
;         Y: Vector containing the y coordinates of the sources. (float if decimal coordinates or not)
;
; KEYWORD PARAMETERS:
;         FPRIOR: Estimation of the flux of the fitted sources
;         (accelerate the fit)
;         MAP_CLEAN: Map cleaned from fitted sources.
;         (useful to check the results of a fit)
;         EXACT: Invert exactly the matrix (do not use conjugate
;         gradient method)
;
;         EFLUX: Vector containing the uncertainties on the output fluxes.
;         COV: Covariance matrix of the output fluxes.
;         Warning: EFLUX and COV suppose a purely instrumental
;         non-correlated noise!
;         The last line and column are associated to mu (the constant
;         background) if mu is not fixed.
;
;         MU: fitted background level.
;         EMU: Uncertainty on mu. 
;         FIXED_MU: if you want to fix manually the background level.
;
; OUTPUTS:
;         FLUX: flux of the fitted sources.
;
; MODIFICATION HISTORY:
; Written by Morgan Cousin in May 2009
; Documentation by Matthieu Bethermin in Feb 2010
; Modified to include compute the uncertainties, to correct of a bug
; in the treatment of the noise_map and to change the normalization of
; the A and B matrices by Matthieu in April 2010 
; Add the possibility to fit the background level by Matthieu in May 2010
;-

PRO fastphot, map, psf, noise_map, X, Y, flux, Fprior = Fprior, map_clean = map_clean, exact = exact, eflux = eflux, cov = cov, fixed_mu = fixed_mu, mu = mu, emu = emu

  if keyword_set(alpha) ne 1 then alpha = 0.

;Compute the size of different elements
  dim1 = (size(map))(1)  	;x size of the map
  dim2 = (size(map))(2)  	;y size of the map
  dim_psf1 = (size(psf))(1) 	;x size of the psf
  dim_psf2 = (size(psf))(2) 	;y size of the psf
  Ns = n_elements(X)            ;Number of sources in the map
  flux = fltarr(Ns)

  if keyword_set(Fprior) ne 1 then Fprior = fltarr(Ns)		;If no prior flux in entrance
  
;Add edge to the map and the noise map
  edge1 = dim_psf1
  edge2 = dim_psf2
  LX = dim1 + 2*edge1 		;Full x size of the map
  LY = dim2 + 2*edge2 		;Full y size of the map
  fullmap = fltarr(LX,LY)
  fullmap(*,*) = -99999. 	;Initialisation for detect bad pixels in the fallowing proress
  map_noise = fltarr(LX,LY)
  fullmap[edge1 : edge1 + dim1 -1, edge2 : edge2 + dim2 -1] = map

  if keyword_set(fixed_mu) then fullmap = fullmap - fixed_mu else PSFOVERNOISE2 = fltarr(Ns)

  ;Add the original map in the full map use for analyse
  map_noise[edge1 : edge1 + dim1 -1, edge2 : edge2 + dim2 -1] = noise_map 	;same treatment for the map_noise
  ;if total(map_noise) eq 0. then begin
  ;  map_noise[*,*] = 1.0
  ;  stop
  ;endif
;Compute the position of the center of the PSF
  c1 = (dim_psf1 -1)/2 	;Center position on x
  c2 = (dim_psf2 -1)/2	;Center position on y
;Shift the sources positions to take into account the added edge
  Xs = X(*) + edge1             ;Warning to the initial shift
  Ys = Y(*) + edge2  
;Compute integer positions
  Xse = round(Xs(*))	;X source position in pixels
  Yse = round(Ys(*))	;Y source position in pixels
;Shift the sources
  dX = Xs(*) - Xse(*)	;Difference between float and integer position
  dY = Ys(*) - Yse(*)
;Construction of the equation 0 = A*Flux + B
;Compute B vector and A matrix
  B = fltarr(Ns)
  A = fltarr(Ns,Ns)
  psfi = fltarr(dim_psf1, dim_psf2) 	;Psf use for i index source
  psfjtot = fltarr(dim_psf1, dim_psf2) 
  psfj = fltarr(dim_psf1, dim_psf2)	;Psf use for j index source
  for ni = 0, Ns -1 do begin   ;Loop on the i sources
                                ;Extract a map cutout around the
                                ;source i

     ;----------------------------------------------------------------
     ;Extracted map around i index source
     sub_map_i = fullmap[Xse(ni) - (dim_psf1-1)/2 : Xse(ni) + (dim_psf1-1)/2, Yse(ni) - (dim_psf2-1)/2 : Yse(ni) + (dim_psf2-1)/2]  		
     ;Extracted map_noise around i index source
     sub_map_noise_i = map_noise[Xse(ni) - (dim_psf1-1)/2 : Xse(ni) + (dim_psf1-1)/2, Yse(ni) - (dim_psf2-1)/2 : Yse(ni) + (dim_psf2-1)/2]
  
     ;----------------------------------------------------------------
     ;Bad pixels treatment and pfsi construction 
     badpix = where(sub_map_i(*,*) lt -99998.) ;Edge map subisation 
           
		      
     psfi(*,*) = 0.								;Initialisation of psfi
     add_sources_interp, psfi, float(c1) + dX(ni), float(c2) + dY(ni), psf, 1. 	;Add of a normalised psf at the i source position 	
    
     if badpix(0) ne -1 then begin
        psfi(badpix) = 0.
     endif
     if badpix(0) ne -1 then begin 
        sub_map_i(badpix) = 0.
     endif
     if badpix(0) ne -1 then begin
        sub_map_noise_i(badpix) = 1.e50
     endif
  
     ;----------------------------------------------------------------
     ;Compute B
     B(ni) = total(-1.*psfi(*,*)*sub_map_i(*,*)/((sub_map_noise_i(*,*))^2))
     if keyword_set(fixed_mu) ne 1 then PSFOVERNOISE2(ni) = total(psfi(*,*)*1./((sub_map_noise_i(*,*))^2))


     for nj = ni, Ns -1 do begin            ;Loop on the j sources

        ;Recovery sources treatment
	;----------------------------------------------------------------
        Si = dim_psf1 - abs(Xse(nj)-Xse(ni)) ;Recovery on i dimension
        Sj = dim_psf2 - abs(Yse(nj)-Yse(ni)) ;Recovery on j dimension

        ;----------------------------------------------------------------
	;Pfsi construction 
        psfjtot(*,*) = 0.
        psfj(*,*) = 0.

        if Si gt 0. and Sj gt 0. then begin   
           add_sources_interp, psfjtot, float(c1) + dX(nj), float(c2) + dY(nj), psf, 1.
           if Xs(nj) ge Xs(ni) then begin
              if Ys(nj) ge Ys(ni) then begin
                 psfj[dim_psf1 - Si :dim_psf1 -1, dim_psf2 - Sj :dim_psf2 -1] = psfjtot[0 : Si -1, 0 : Sj -1]
              endif else begin 
                 psfj[dim_psf1 - Si :dim_psf1 -1, 0 : Sj -1] = psfjtot[0 : Si -1, dim_psf2 - Sj :dim_psf2 -1]
              endelse   
           endif else begin
              if Ys(nj) ge Ys(ni) then begin
                 psfj[0 : Si -1, dim_psf2 - Sj :dim_psf2 -1] = psfjtot[dim_psf1 - Si : dim_psf1 -1, 0 : Sj -1]
              endif else begin 
                 psfj[0 : Si -1 ,0 : Sj -1] = psfjtot[dim_psf1 - Si : dim_psf1 -1 ,dim_psf2 - Sj :dim_psf2 -1]
              endelse
           endelse
  
	   ;----------------------------------------------------------------
           ;Compute A
           A(ni,nj) = total(psfi(*,*)*psfj(*,*)*1./((sub_map_noise_i(*,*))^2))
           A(nj,ni) = A(ni,nj)  ;Compute the other half of the A matrix (symetrical)

        endif
     endfor
  endfor
  B(*) = -B(*)




;Add the background minimisation to the fit

if keyword_set(fixed_mu) ne 1 then begin
Nmatrix = n_elements(B)
Aold = A
Bold = B
A = fltarr(Nmatrix+1,Nmatrix+1)
A[0:Nmatrix-1,0:Nmatrix-1] = Aold
B = fltarr(Nmatrix+1)
B[0:Nmatrix-1] = Bold 

sel = where(fullmap gt -99998. and noise_map ne 0.)

A(Nmatrix,Nmatrix) = total(1./noise_map(sel)^2)
B(Nmatrix) = total(fullmap(sel)*1./noise_map(sel)^2)

A(0:Nmatrix-1,Nmatrix) = PSFOVERNOISE2
A(Nmatrix,0:Nmatrix-1) = PSFOVERNOISE2

endif




;Choose the method to resolve the equation
;----------------------------------------------------------------

if keyword_set(eflux) then eflux = B*0.

  if Ns lt 3 or keyword_set(exact) or keyword_set(eflux) or keyword_set(cov) then begin                        
     Ainv = invert(A)				;Extact resolution of the linear equation
     flux = Ainv#B
     
     cov = Ainv
     for k = 0, N_elements(B)-1 do begin
        eflux(k) = sqrt(cov(k,k))
     endfor

  endif else begin

     if keyword_set(fixed_mu) then X0 = Fprior(*) else X0 = [Fprior,0.]

     fastphot_gradsolver, A, B, X0 = X0, flux    	;Resolution by the conjugate gradient method  
  endelse

  if keyword_set(fixed_mu) ne 1 then begin
     mu = flux(Ns)
     if keyword_set(eflux) then emu = eflux(Ns)
     flux = flux[0:Ns-1]
     if keyword_set(eflux) then eflux = eflux[0:Ns-1]
  endif
  
;Clean map construction
;----------------------------------------------------------------
  if arg_present(map_clean) then begin
     map_clean = map		
     add_sources_interp, map_clean, X, Y, psf, flux*(-1.) 	
  endif

END
